import enum
import typing
from collections.abc import Iterator, Sequence
from typing import Literal, NewType, final, overload

import attr
import cattr
from arrow import Arrow
from cattr import Converter, override
from cattrs.gen._consts import AttributeOverride

# Revert this commit when cattrs 24.x drops.
# type ProjectType = Literal["mod", "modpack", "resourcepack", "shader"]
# type ModSide = Literal["unknown", "required", "optional", "unsupported"]


class ModSideValue(enum.Enum):  # noqa: D101
    UNKNOWN = "unknown"
    REQUIRED = "required"
    OPTIONAL = "optional"
    UNSUPPORTED = "unsupported"


# deliberately undocumented because i'm lazy. see the modrinth docs.
# see client.py for why we need some hackery with cattrs overrides.
# duck typing kinda works at least...

ProjectId = NewType("ProjectId", str)
VersionId = NewType("VersionId", str)


@attr.s(frozen=True, kw_only=True)
class ProjectInfoMixin:
    """
    Common fields between :class:`.ProjectInfoFromSearch` and :class:`.ProjectInfoFromProject`.
    """

    @classmethod
    def configure_converter(cls, converter: Converter) -> None:  # noqa: D102
        # some renames are needed because modrinth is really fucking stupid
        # and likes

        for klass in (ProjectInfoFromSearch, ProjectInfoFromProject):
            converter.register_structure_hook(
                klass,
                cattr.gen.make_dict_structure_fn(
                    klass,
                    converter,
                    raw_categories=override(rename="categories"),
                    raw_versions=override(rename="versions"),
                    **klass.get_overrides(),  # type: ignore  # ok, pyright
                ),
            )

    #: The ID of this project.
    id: ProjectId

    #: The project's URL slug.
    slug: str = attr.ib()
    #: The human-readable title for the project.
    title: str = attr.ib()
    #: Project's short description.
    description: str = attr.ib()  # short description

    # really terrible way of doing it, imo.
    client_side: ModSideValue = attr.ib()
    server_side: ModSideValue = attr.ib()

    raw_categories: list[str] = attr.ib()
    raw_versions: list[str] = attr.ib()


@attr.s(slots=True, frozen=True, kw_only=True)
@final
class ProjectInfoFromSearch(ProjectInfoMixin):
    """
    Wrapper type for project information returned from ``/search``.
    """

    @classmethod
    def get_overrides(cls) -> dict[str, AttributeOverride]:  # noqa: D102
        return {
            "id": override(rename="project_id"),
        }

    id: ProjectId = attr.ib()  # base62 encoded, but we treat it as an opaque string

    @property
    def game_versions(self) -> list[str]:
        """
        Gets the game versions this project is for.
        """

        return self.raw_versions


@attr.s(slots=True, frozen=True, kw_only=True)
@final
class ProjectInfoFromProject(ProjectInfoMixin):
    """
    Wrapper type for project information returned from ``/project``.
    """

    @classmethod
    def get_overrides(cls) -> dict[str, AttributeOverride]:  # noqa: D102
        return {}

    id: ProjectId = attr.ib()

    #: The game versions this project is for.
    game_versions: list[str] = attr.ib()


@attr.s(slots=True, frozen=True, kw_only=True)
@final
class ProjectSearchResult(Sequence[ProjectInfoFromSearch]):
    """
    Wrapper type for the result of a project search.
    """

    # wtf is this name
    hits: list[ProjectInfoFromSearch] = attr.ib()
    offset: int = attr.ib()
    limit: int = attr.ib()
    total_hits: int = attr.ib()

    def __bool__(self) -> bool:
        return bool(self.hits)

    @typing.override
    def __iter__(self) -> Iterator[ProjectInfoFromSearch]:
        return iter(self.hits)

    @typing.override
    def __len__(self) -> int:
        return len(self.hits)

    @overload
    def __getitem__(self, index: int) -> ProjectInfoFromSearch:
        ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[ProjectInfoFromSearch]:
        ...

    @typing.override
    def __getitem__(
        self,
        index: int | slice,
    ) -> ProjectInfoFromSearch | Sequence[ProjectInfoFromSearch]:
        return self.hits[index]


@attr.s(slots=True, frozen=True, kw_only=True)
@final
class ProjectVersionFile:
    """
    A single file for a project version.
    """

    #: The URL for this version.
    url: str = attr.ib()

    #: The real filename for this version.
    filename: str = attr.ib()

    #: If this version is the *primary* file or not (i.e. the main mod file).
    primary: bool = attr.ib(default=False)

    #: The size of the file, in bytes.
    size: int = attr.ib()

    #: Opaque set of hashes in unspecified algorithms.
    hashes: dict[str, str] = attr.ib()


@attr.s(slots=True, frozen=True, kw_only=True)
@final
class ProjectVersionRelation:
    """
    A single relation within a project version.
    """

    #: The ID of the other project version in this relationship.
    version_id: VersionId | None = attr.ib()

    #: The ID of the other project in this relationship.
    project_id: ProjectId = attr.ib()

    #: The type of this relationship.
    dependency_type: Literal["required", "optional", "incompatible", "embedded"] = attr.ib()


@attr.s(slots=True, frozen=True, kw_only=True)
@final
class ProjectVersion:
    """
    A single version within a project.
    """

    @classmethod
    def configure_converter(cls, converter: Converter) -> None:  # noqa: D102
        converter.register_structure_hook(
            ProjectVersion,
            cattr.gen.make_dict_structure_fn(
                ProjectVersion, converter, relationships=override(rename="dependencies")
            ),
        )

    #: The ID for this version.
    id: VersionId = attr.ib()

    #: The project ID for this version.
    project_id: ProjectId = attr.ib()

    #: The name of this version.
    name: str = attr.ib()

    #: The number for this version.
    version_number: str = attr.ib()

    #: The game versions for this mod.
    game_versions: list[str] = attr.ib()

    #: If this version is featured or not.
    featured: bool = attr.ib()

    #: The list of modloaders this version is for.
    loaders: list[str] = attr.ib()

    #: When this version was uploaded.
    date_published: Arrow = attr.ib()

    #: The stability status for this version.
    version_type: Literal["release", "beta", "alpha"] = attr.ib()

    #: The relationships for this version.
    relationships: list[ProjectVersionRelation] = attr.ib(factory=list)

    #: The list of :class:`.ProjectVersionFile` instances for this single version (e.g. main mod
    #: and sources jar, for weirdos who use Modrinth maven).
    files: list[ProjectVersionFile] = attr.ib()

    def __attrs_post_init__(self) -> None:
        if len(self.files) == 1:
            object.__setattr__(self.files[0], "primary", True)

    @property
    def primary_file(self) -> ProjectVersionFile:
        """
        Gets the primary file for this project.
        """

        for file in self.files:
            if file.primary:
                return file

        raise ValueError("modrinth fucked up, missing primary version")



import argparse
import contextlib
import pprint
import sys
from pathlib import Path

import cattrs
import httpx
import platformdirs
import tomlkit

from kamuidrome.cache import ModCache
from kamuidrome.cli.add import add_mod_by_project_id, add_mod_by_searching, add_mod_by_version_id
from kamuidrome.cli.init import interactively_create_pack
from kamuidrome.cli.list import list_indexed_mods
from kamuidrome.cli.update import download_all_mods, update_all_mods
from kamuidrome.meta import LocalMetadata
from kamuidrome.modrinth.client import ModrinthApi
from kamuidrome.modrinth.models import ProjectId, VersionId
from kamuidrome.modrinth.mrpack import create_mrpack
from kamuidrome.pack import load_local_pack


def main() -> int:
    """
    Main entrypoint.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--pack-dir",
        help="The directory to use (defaults to CWD)",
        default=Path.cwd(),
        type=Path,
    )
    parser.add_argument(
        "--cache-dir",
        help="The cache directory to use (defaults to ~/.cache)",
        default=Path(platformdirs.user_cache_dir("kamuidrone")),
        type=Path,
    )

    subcommands = parser.add_subparsers(
        title="Subcommands",
        description="Pack management actions",
        dest="subcommand",
        required=True,
    )

    init_group = subcommands.add_parser("init", help="Initialise a new modpack interactively")
    init_group.add_argument(
        "--git", help="Create a Git repository and .gitignore", action="store_true", default=False
    )

    add_mod = subcommands.add_parser("add", help="Adds a new mod")
    add_mod.add_argument(
        "--always-select",
        help="Always shows a selection prompt for search",
        action="store_true",
        default=False,
    )
    add_group = add_mod.add_mutually_exclusive_group(required=True)
    add_group.add_argument(
        "-s", "--search", help="Adds a mod by searching for the specified argument", default=None
    )
    add_group.add_argument(
        "-p", "--project-id", help="Adds a mod using the specified project ID", default=None
    )
    add_group.add_argument(
        "-V", "--version-id", help="Adds a mod using the specified version ID", default=None
    )

    deploy_group = subcommands.add_parser("deploy", help="Deploys a modpack")
    deploy_group_args = deploy_group.add_mutually_exclusive_group(required=False)
    deploy_group_args.add_argument(
        "-i", "--instance", help="The name of the Prism instance to deploy to", default=None
    )
    deploy_group_args.add_argument(
        "-d", "--directory", help="The path of the directory to deploy to", default=None, type=Path
    )

    pin_group = subcommands.add_parser(name="pin", help="Pins a mod version to the current version")
    pin_group.add_argument("MOD", nargs="+", help="The mod name or ID to pin")

    subcommands.add_parser(name="list", help="List indexed mods")

    subcommands.add_parser(name="download", help="Downloads all mods in the index")
    subcommands.add_parser(name="update", help="Updates all mods and dependenciess in the index")

    export_group = subcommands.add_parser("export", help="Exports pack as an mrpack file")
    export_group.add_argument(
        "FILENAME", help="The name of the file or directory to write", nargs="?", default=None
    )
    export_group.add_argument(
        "--ci-mode",
        action="store_true",
        default=False,
        help="Outputs an unpacked mrpack instead of a fully packed one",
    )
    export_group.add_argument(
        "--server-only", action="store_true", default=False, help="Only outputs server-side mods"
    )

    args = parser.parse_args()
    cache_dir: Path = args.cache_dir
    cache = ModCache(cache_dir=cache_dir)

    subcommand: str = args.subcommand

    if subcommand == "init":
        return interactively_create_pack(args.pack_dir, with_git=args.git)

    pack = load_local_pack(args.pack_dir)

    with httpx.Client() as client:
        api = ModrinthApi(client)

        if subcommand == "add":
            search_query: str | None = args.search
            if search_query is not None:
                return add_mod_by_searching(pack, api, cache, search_query, args.always_select)

            project_id: str | None = args.project_id
            if project_id is not None:
                return add_mod_by_project_id(pack, api, cache, ProjectId(project_id))

            version_id: str = args.version_id
            add_mod_by_version_id(pack, api, cache, VersionId(version_id))

        elif subcommand == "deploy":
            instance_name: str | None = args.instance
            folder_name: Path | None = args.directory

            local_metadata: LocalMetadata | None = None

            with contextlib.suppress(FileNotFoundError, KeyError):
                localpack = pack.directory / "localpack.toml"
                data = tomlkit.loads(localpack.read_text())
                local_metadata = cattrs.structure(data, LocalMetadata)

            if folder_name is not None:
                return pack.deploy_to_directory(cache, folder_name, local_metadata)

            if instance_name is None:
                if local_metadata is None:
                    parser.error(
                        "expected either a folder name, an instance name, "
                        "or for the instance name to be set in localpack.toml"
                    )

                instance_name = local_metadata.instance_name

            return pack.deploy_to_instance(cache, instance_name, local_metadata)

        elif subcommand == "pin":
            return pack.pin(" ".join(args.MOD))

        elif subcommand == "download":
            return download_all_mods(pack, api, cache)

        elif subcommand == "update":
            return update_all_mods(pack, api, cache)

        elif subcommand == "list":
            return list_indexed_mods(pack)

        elif subcommand == "export":
            export_name: str | None = args.FILENAME
            ci_mode: bool = args.ci_mode
            server_only: bool = args.server_only

            if export_name is None:
                if ci_mode:
                    parser.error("must provide a directory name in ci mode")
                else:
                    export_path = (Path.cwd() / pack.metadata.name).with_suffix(".mrpack")
            elif not ci_mode:
                export_path = Path(export_name).with_suffix(".mrpack")
            else:
                export_path = Path(export_name)
                export_path.mkdir(exist_ok=True, parents=True)

            create_mrpack(pack, api, export_path, ci_mode=ci_mode, server_only=server_only)
            return 0

    return 0


if __name__ == "__main__":

    def _run():
        try:
            sys.exit(main())
        except httpx.HTTPStatusError as e:
            pprint.pprint(e.response.json())

    _run()
